//
// Created by sofie on 16.11.22.
//
#include <vector>
#include "sensor_msgs/PointCloud2.h"
#include "sensor_msgs/point_cloud2_iterator.h"
#include <cmath>
//pcl:
#include <pcl_conversions/pcl_conversions.h> 	//convert msg Pointcloud to pcl PointCloud
#include <pcl/point_types.h>
#include <pcl/PCLPointCloud2.h>
#include <pcl/conversions.h>
#include <pcl_ros/transforms.h>
#include <pcl/kdtree/kdtree_flann.h>

#ifndef ROCK_ANALYSIS_INTENSITYPOINT_H
#define ROCK_ANALYSIS_INTENSITYPOINT_H


/**
 * Class for Keypoint(area), looking at the neighborhood to get a mean intensity.
 * Intensities of Neighbour-points get weighted into one intensity that is presumed to have the
 * distance and incidence angle of the center-point.
**/
class IntensityPoint {

private:
    
    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud;
    pcl::PointXYZI keypoint;              //set keypoint/center at initialization of class, ignore intensity attribute (XYZI is needed for radius search)
    double distance = 0;                  //calculated as vector size from scanner to keypoint
    std::vector<int> roiIdx;              //region of interest indices in pointcloud (radius around keypoint, including keypoint)
    std::vector<double> meanintensity;    //mean with standard deviation
    std::vector<double> roiI;             //Intensities in region of interest, includes keypoint
    float r = 0.02;                       //radius for kdtree search, default 0.2m 
    double treshold = 0.01;               //treshold for distance check unit meter

public:
    /**
    *   simplest constructor. Parameters to give: coordinates of point and cloud in which the point is found
    **/
    IntensityPoint(float x, float y, float z, pcl::PointCloud<pcl::PointXYZI>::Ptr pcl){
        this->keypoint.x = x;
        this->keypoint.y = y;
        this->keypoint.z = z;
        this->cloud = pcl;

        //calculate distance
        this->distance = sqrt(pow(x,2) + pow(y,2) + pow(z,2)); //if static, (x0,y0,z0) = (0,0,0)  //determine actual position if non static

        //get roiIdx
        calculateNeighbours();

        //fill roiI with intensities of neighbours
        fillRoiI();

        //calculate mean intensity, needs roiI
        calculateMeanI();
    }

    /**
    *   radius&treshold constructor. Parameters to give: coordinates of point and cloud in which the point is found as well as radius to be
    *   used in neighbour search and treshold for distance delete
    **/
    IntensityPoint(pcl::PointXYZ keypoint, pcl::PointCloud<pcl::PointXYZI>::Ptr pcl, float radius, double treshold){
        this->keypoint.x = keypoint.x;
        this->keypoint.y = keypoint.y;
        this->keypoint.z = keypoint.z;
        this->cloud = pcl;
        this->r = radius;
        this->treshold = treshold;

        //calculate distance
        this->distance = sqrt(pow(keypoint.x,2) + pow(keypoint.y,2) + pow(keypoint.z,2)); //if static, (x0,y0,z0) = (0,0,0)  // determine actual position if non static

        //calculate roiIdx
        calculateNeighbours();

        //fill roiI with intensities of neighbours, needs roiIdx
        fillRoiI();

        //calculate mean intensity, needs roiI
        calculateMeanI();
    }

 /**
    *   radius&treshold constructor. Parameters to give: coordinates of point as Eigen::Vector3d and cloud in which the point is found as well as radius to be
    *   used in neighbour search and treshold for distance delete
    **/
    IntensityPoint(Eigen::Vector3f keypoint, pcl::PointCloud<pcl::PointXYZI>::Ptr pcl, float radius, double treshold){
        this->keypoint.x = keypoint[0];
        this->keypoint.y = keypoint[1];
        this->keypoint.z = keypoint[2];
        this->cloud = pcl;
        this->r = radius;
        this->treshold = treshold;

        //calculate distance
        this->distance = sqrt(pow(keypoint[0],2) + pow(keypoint[1],2) + pow(keypoint[2],2)); //if static, (x0,y0,z0) = (0,0,0)  //determine actual position if non static

        //calculate roiIdx
        calculateNeighbours();

        //fill roiI with intensities of neighbours, needs roiIdx
        fillRoiI();

        //calculate mean intensity, needs roiI
        calculateMeanI();
    }



    /**
    *   constructor for calibration. Parameters to give:  keypoint and cloud in which the point is found as well as a radius for kdtree search
    **/
    IntensityPoint(pcl::PointXYZ keypoint, pcl::PointCloud<pcl::PointXYZI>::Ptr inputpcl, float radius){
        this->keypoint.x = keypoint.x;
        this->keypoint.y = keypoint.y;
        this->keypoint.z = keypoint.z;
        this->cloud = inputpcl;
        this->r = radius;

        //calculate distance
        this->distance = sqrt(pow(keypoint.x,2) + pow(keypoint.y,2) + pow(keypoint.z,2)); //if (x0,y0,z0) = (0,0,0)  //determine actual position if non static

        //calculate roiIdx
        calculateNeighbours();

        //fill roiI with intensities of neighbours
        fillRoiI();

        //calculate mean intensity, needs roiI
        calculateMeanI();
    }

    /**
    *   functions to get parameters
    **/
    std::vector<double> getMeanI(){ return this->meanintensity; }
    double getDistance(){ return this->distance; }


private:
    void fillRoiI(){

        roiI.resize(roiIdx.size());

        for (int i = 0; i < roiIdx.size(); ++i) { roiI[i] = cloud->points[roiIdx[i]].intensity;}

    }

    void calculateNeighbours(){

        pcl::KdTreeFLANN<pcl::PointXYZI> kdtree;
        kdtree.setInputCloud(this->cloud);
       
        //------nearest neighbour points search within Radius------
        
        std::vector<float> neighbourDistancesSquared;
        std::vector<int> neighbourIdx;
    
        float radius = this->r; 
        kdtree.radiusSearch (this->keypoint, radius, neighbourIdx, neighbourDistancesSquared);
        this->roiIdx = neighbourIdx;
       
        ROS_INFO("neighbourIDX size = %ld", roiIdx.size());


    }

    /**
    *   Arithmetic mean
    **/
    double getArithMean(std::vector<double> values) {

        double sum = 0; 

        for (int i = 0; i < values.size(); ++i) { sum += values[i]; }

        return sum / values.size();
    }

   /**
    *   standard deviation
   **/
    double getStandardDev(std::vector<double> values) {

        double mean = getArithMean(values);
        double sum = 0; 

        for (int i = 0; i < values.size(); ++i) { sum += (values[i] - mean)*(values[i] - mean);  }

        return sqrt(sum/(values.size()-1));

    }

    /*
     * get distances for roi points
     */
    std::vector<double> getRoiDistances(){

        std::vector<double> distances(roiIdx.size());

        //fill distances with roi distances
        for (int i = 0; i < roiIdx.size(); i++) { distances[i] = sqrt(pow(this->cloud->points[roiIdx[i]].x, 2)+pow(this->cloud->points[roiIdx[i]].y, 2)+pow(this->cloud->points[roiIdx[i]].z, 2)); }  //distance to scanner 

        return distances;

    }

    /**
    *   Is the measurement valid: does it lie inside of treshold
    *   additionally check for nan (e.g. points the laserscanner couldnt detect cause of range)
    **/
    bool isValid(double value, double Uptreshold, double Downtreshold){

        if (isnan(value) or !isfinite(value) or value < Downtreshold  or value > Uptreshold){ return false; }
        else { return true; }

    }



    void distanceCheck(){

        std::vector<int> tempIdx  = roiIdx;
        std::vector<double> tempI = roiI;


    //------------------------Remove NonObject Measurements-----------------------------

        //remove if too far away or close, meaning not the target Object anymore
        int i = 0;
        std::vector<double> distances = getRoiDistances();

        while(i < tempI.size()){

            //remove invalid value from roiI and roiIdx
            if (abs(distance - distances[i]) < treshold){
                ++i;
            } else {
                distances.erase(distances.begin()+i);
                tempI.erase(tempI.begin()+i);
                tempIdx.erase(tempIdx.begin()+i);
            }
            
        }

        if(tempIdx.size() == 0){
            ROS_ERROR("[IntensityPoint] distance check failed, proceeding without");
        } else {
            roiI = tempI;
            roiIdx = tempIdx;
        }
        
    }

    void sigmaband(){

    //------------------------------Sigma-interval-----------------------------------
    //Is the measurement valid: does it lie inside of 1.96*deviation (Normal distribution, 95% inside)

        double mean = getArithMean(roiI);
        double s = getStandardDev(roiI);
        double boundD = mean - 1.96*s;
        double boundU = mean + 1.96*s;

        //iterate through all values, removing the invalid ones: being everything deviating more than 1.96*s from the mean
        std::vector<int> tempIdx  = roiIdx;
        std::vector<double> tempI = roiI;
        int i = 0;

        while(i < tempI.size()){

            if (isValid(tempI[i], boundU, boundD)){ 
                    ++i;
            } else {//remove invalid values from roiI and roiIdx
                
                tempI.erase(tempI.begin()+i);
                tempIdx.erase(tempIdx.begin()+i);
                mean = getArithMean(tempI);  //recompute mean and deviance and start anew 
                s = getStandardDev(tempI);
                boundD = mean - 1.96*s;
                boundU = mean + 1.96*s;
                i = 0;
            }
            
        } 

        if(tempI.size() == 0){
            ROS_ERROR("[IntensityPoint] sigmaband failed, proceeding without");
            mean = getArithMean(roiI);  
            s = getStandardDev(roiI);
        } else {
            roiI = tempI;
            ROS_INFO("[IntensityPoint] sigmaband finished roi size: %ld", roiI.size());
        }

        meanintensity[0] = mean;
        meanintensity[1] = s;

    }

    /**
    *   First check if still target Object (distance check) then get rid of unrealistic measurements (sigmaband). 
    *   Assumes normal distribution.
    *   Then set meanI and standard deviation for class Object (done by sigmaband).
    **/
    void calculateMeanI(){

        distanceCheck();

        meanintensity.resize(2);
        meanintensity[0] = getArithMean(roiI);
        meanintensity[1] = getStandardDev(roiI);
        
        sigmaband();      
       
    }


};// end of Point class


#endif //ROCK_ANALYSIS_INTENSITYPOINT_H
