//
// Created by sofie on 24.11.22.
// run with: $launch rock_analysis calibration.launch
//
#include <iostream>
#include <pcl/io/pcd_io.h>
#include <pcl/point_types.h>
#include "rock_analysis/tools/intensityPoint.h"
#include "rock_analysis/tools/calibration.h"
#include <rock_analysis/tools/regression.h>
#include "pcl_ros/point_cloud.h"
#include <ros/ros.h>
#include "sensor_msgs/point_cloud_conversion.h" //for converting PointCloud to PointCloud2

using namespace Eigen;
const static IOFormat CSVFormat(StreamPrecision, DontAlignCols, ", ", "\n");

class CalibrationSet{

private:
//create node handle
ros::NodeHandle handle;

//pointclouds and values
pcl::PointCloud<pcl::PointXYZI> livoxPcl;
int livoxCnt; //counts how many livox topics are merged in livoxPcl
pcl::PointCloud<pcl::PointXYZI> rieglPcl;
bool rieglreceived;
pcl::PointCloud<pcl::PointXYZI> ousterPcl;
int ousterCnt;
std::string file;  
std::vector<Eigen::Vector3f> l = std::vector<Eigen::Vector3f>(4);
std::vector<Eigen::Vector3f> r = std::vector<Eigen::Vector3f>(4);
std::vector<Eigen::Vector3f> o = std::vector<Eigen::Vector3f>(4);

//regression
regression rieglReg, ousterReg, livoxReg;

//subscriber
ros::Subscriber livoxsub;
ros::Subscriber rieglsub;
ros::Subscriber oustersub;

public:
CalibrationSet(){

    //get rosparams
    handle.getParam("calpath", file);

    //get needed calibration points from csv file
    getPoints();

    //subscribe to scanner topics
    livoxCnt = 0;
    ousterCnt = 0;
    rieglreceived = false;
    livoxsub = handle.subscribe("livox/lidar", 1000, &CalibrationSet::livoxCallback, this);
    rieglsub = handle.subscribe("/riegl3d", 1000, &CalibrationSet::rieglCallback, this);
    oustersub = handle.subscribe("/ouster/points", 1000, &CalibrationSet::ousterCallback, this);
   
    ROS_INFO("[Calibration] subscribed to livox/lidar /riegl3d /ouster/points");
    ROS_INFO("Waiting to receive riegl scan.");

}

private:
/*
 * takes points from rosparams 
 */
void getPoints(){

    ROS_INFO("[Calibration] getting points.");
    Eigen::Vector3f point;
    std::string pointstr;
    
    //fill livox points
    handle.getParam("l99", pointstr);
    std::istringstream lstrm(pointstr);
    lstrm >> point[0] >> point[1] >> point[2];
    l[0] = point;
    handle.getParam("l75", pointstr);
    std::istringstream l1strm(pointstr);
    l1strm >> point[0] >> point[1] >> point[2];
    l[1] = point;
    handle.getParam("l50", pointstr);
    std::istringstream l2strm(pointstr);
    l2strm >> point[0] >> point[1] >> point[2];
    l[2] = point;
    handle.getParam("l2", pointstr);
    std::istringstream l3strm(pointstr);
    l3strm >> point[0] >> point[1] >> point[2];
    l[3] = point;

    //fill riegl points
    handle.getParam("r99", pointstr);
    std::istringstream rstrm(pointstr);
    rstrm >> point[0] >> point[1] >> point[2];
    r[0] = point;
    handle.getParam("r75", pointstr);
    std::istringstream r1strm(pointstr);
    r1strm >> point[0] >> point[1] >> point[2];
    r[1] = point;
    handle.getParam("r50", pointstr);
    std::istringstream r2strm(pointstr);
    r2strm >> point[0] >> point[1] >> point[2];
    r[2] = point;
    handle.getParam("r2", pointstr);
    std::istringstream r3strm(pointstr);
    r3strm >> point[0] >> point[1] >> point[2];
    r[3] = point;

    //fill ouster points
    handle.getParam("o99", pointstr);
    std::istringstream ostrm(pointstr);
    ostrm >> point[0] >> point[1] >> point[2];
    o[0] = point;
    handle.getParam("o75", pointstr);
    std::istringstream o1strm(pointstr);
    o1strm >> point[0] >> point[1] >> point[2];
    o[1] = point;
    handle.getParam("o50", pointstr);
    std::istringstream o2strm(pointstr);
    o2strm >> point[0] >> point[1] >> point[2];
    o[2] = point;
    handle.getParam("o2", pointstr);
    std::istringstream o3strm(pointstr);
    o3strm >> point[0] >> point[1] >> point[2];
    o[3] = point;
}

/*
 *  simple method to write calibration data (max and min intensity) to <scanner>.csv in calpath (rosparam)
*/
void write_csv(std::string scanner, std::vector<double> data){ 

   std::string filename = file + "/" + scanner + ".csv";
   ROS_INFO("writing to %s ", filename.c_str());
   //create an output filestream object
   std::ofstream file(filename, std::ios::out);

   //write intensity data to file
   file << scanner << "\n";
   file << data[0] << "\n";
   file << data[1] << "\n"; 
    
   //ROS_INFO("Written data: [%f,%f]", data[0], data[1]);

   file.close();

}

/*
 *  simple method to use IntensityPoint object for calculating Intensity of target at keypoint in PointCloud pcl
 */
std::vector<double> calculate(Eigen::Vector3f keypoint, pcl::PointCloud<pcl::PointXYZI> pcl){
    
    IntensityPoint CalibrationTarget(keypoint, pcl.makeShared(), 0.015, 0.01); //one spectralon has diameter 0.038 -> radius = 0.019, using 0.015, I wont have selected true centre, distance treshold is 1cm
    return CalibrationTarget.getMeanI();

}

void calculations(){

    std::vector<double> range(4);
    std::vector<double> meanIwhite, meanIdarkgrey, meanIgrey, meanIblack;
   
    //--------------livox-------------------
    
    //white spectralon 99%
    meanIwhite = calculate(l[0], livoxPcl); //reusing old variables
    ROS_INFO("Calculated livox 99%% calibration intensity of [%lf+-%lf]", meanIwhite[0], meanIwhite[1]);
    //grey spectralon 75%
    meanIgrey = calculate(l[1], livoxPcl);
    ROS_INFO("Calculated livox 75%% calibration intensity of [%lf+-%lf]", meanIgrey[0], meanIgrey[1]);
    //grey spectralon 50%
    meanIdarkgrey = calculate(l[2], livoxPcl);
    ROS_INFO("Calculated livox 50%% calibration intensity of [%lf+-%lf]", meanIdarkgrey[0], meanIdarkgrey[1]);
    //black spectralon 2%
    meanIblack = calculate(l[3], livoxPcl);
    ROS_INFO("Calculated livox 2%% calibration intensity of [%lf+-%lf]", meanIblack[0], meanIblack[1]);
    //write to csv 
    std::vector<double> x = {0.2, 0.5, 0.75, 0.99};
    std::vector<double> ldata = {meanIblack[0], meanIdarkgrey[0], meanIgrey[0], meanIwhite[0]};

    livoxReg.setParams(x, ldata);
    double maxI = livoxReg.predict(1);
    double minI = livoxReg.predict(0);
    
    range[0] = maxI;
    range[1] = minI;
    write_csv("livox", range);
    
    //--------------riegl-------------------
        
    //white spectralon 99%
    meanIwhite = calculate(r[0], rieglPcl); //reusing old variables
    ROS_INFO("Calculated riegl 99%% calibration intensity of [%lf+-%lf]", meanIwhite[0], meanIwhite[1]);
    //grey spectralon 75%
    meanIgrey = calculate(r[1], rieglPcl);
    ROS_INFO("Calculated riegl 75%% calibration intensity of [%lf+-%lf]", meanIgrey[0], meanIgrey[1]);
    //grey spectralon 50%
    meanIdarkgrey = calculate(r[2], rieglPcl);
    ROS_INFO("Calculated riegl 50%% calibration intensity of [%lf+-%lf]", meanIdarkgrey[0], meanIdarkgrey[1]);
    //black spectralon 2%
    meanIblack = calculate(r[3], rieglPcl);
    ROS_INFO("Calculated riegl 2%% calibration intensity of [%lf+-%lf]", meanIblack[0], meanIblack[1]);
    //write to csv 
    std::vector<double> rdata = {meanIblack[0], meanIdarkgrey[0], meanIgrey[0], meanIwhite[0]};
    
    rieglReg.setParams(x, rdata);
    maxI = rieglReg.predict(1);
    minI = rieglReg.predict(0);
    
    range[0] = maxI;
    range[1] = minI;
    write_csv("riegl", range);
    

    //--------------ouster-----------------
    
    //white spectralon 99%
    meanIwhite = calculate(o[0], ousterPcl); //reusing old variables
    ROS_INFO("Calculated ouster 99%% calibration intensity of [%lf+-%lf]", meanIwhite[0], meanIwhite[1]);
    //grey spectralon 75%
    meanIgrey = calculate(o[1], ousterPcl);
    ROS_INFO("Calculated ouster 75%% calibration intensity of [%lf+-%lf]", meanIgrey[0], meanIgrey[1]);
    //grey spectralon 50%
    meanIdarkgrey = calculate(o[2], ousterPcl);
    ROS_INFO("Calculated ouster 50%% calibration intensity of [%lf+-%lf]", meanIdarkgrey[0], meanIdarkgrey[1]);
    //black spectralon 2%
    meanIblack = calculate(o[3], ousterPcl);
    ROS_INFO("Calculated ouster 2%% calibration intensity of [%lf+-%lf]", meanIblack[0], meanIblack[1]);
    //write to csv 
    std::vector<double> odata = {meanIblack[0], meanIdarkgrey[0], meanIgrey[0], meanIwhite[0]};

    ousterReg.setParams(x, odata);
    maxI = ousterReg.predict(1);
    minI = ousterReg.predict(0);
    
    range[0] = maxI;
    range[1] = minI;
    write_csv("ouster", range);

    ROS_INFO("Killing...");
    ros::shutdown();  //finished: kill node
}

/*
 *  livox Callback, also checks if other pointclouds are aquired already and starts calculation if so
 */
void livoxCallback(const pcl::PointCloud<pcl::PointXYZI>::ConstPtr& pcl){ 
    //add incoming points to excisting pointcloud
    livoxPcl += *pcl;
    livoxCnt += 1;

    //if riegl pointcloud is already aquired and others are big enough 
    //(at least 50 messages, should be more due to riegl scan time) start calculation
    if (rieglreceived && livoxCnt >= 50 && ousterCnt >= 50){ calculations();} 

}

/*
 *  riegl Callback, gets riegl pointcloud
 */
void rieglCallback(const sensor_msgs::PointCloudConstPtr& msg){
    ROS_WARN("RIEGL RECEIVED");
    //converting incoming PointCloud to pcl pointcloud
   sensor_msgs::PointCloud2 temp;
    if (!sensor_msgs::convertPointCloudToPointCloud2 (*msg, temp))
      {
        ROS_ERROR ("[IntensityCorrection] Conversion from sensor_msgs::PointCloud to sensor_msgs::PointCloud2 failed!");
        return;
      }
   pcl::PCLPointCloud2 pcl_pc2;
   pcl_conversions::toPCL(temp,pcl_pc2);
   pcl::PointCloud<pcl::PointXYZI> pcl;
   pcl::fromPCLPointCloud2(pcl_pc2, pcl);

   rieglPcl = pcl;
   rieglreceived = true;
}

/*
 *  ouster Callback
 */
void ousterCallback(const sensor_msgs::PointCloud2ConstPtr& msg){ 
    //transform incoming point cloud to pcl point cloud with point type XYZI
    pcl::PCLPointCloud2 pcl_pc2;
    pcl_conversions::toPCL(*msg,pcl_pc2);
    pcl::PointCloud<pcl::PointXYZI> OusterPcl;
    pcl::fromPCLPointCloud2(pcl_pc2,OusterPcl);

    //add incoming points to excisting pointcloud
    ousterPcl += OusterPcl;
    ousterCnt += 1;
}

};


int main(int argc, char *argv[])
{

    //initialize Ros node
    ros::init(argc, argv, "calibration");

    ROS_WARN("Calibration running...");
    CalibrationSet set;

    ros::spin();

}