#include <ros/ros.h>
#include "std_msgs/Float32.h"
#include "rock_analysis/tools/intensityPoint.h"
#include "rock_analysis/tools/calibration.h"
#include "pcl_ros/point_cloud.h"
#include <fstream>
#include <ctime>
//to make directory
#include <sys/stat.h>
#include <sys/types.h>
//for clicked point message
#include <geometry_msgs/PointStamped.h>

#include "sensor_msgs/point_cloud_conversion.h" //for converting PointCloud to PointCloud2

//define wavelenghts of scanners 
enum : int {l_lambda = 905, o_lambda = 865, r_lambda = 1550};


class RockIntensity
{
private:
	//create node handle
 	ros::NodeHandle n;

   //pointclouds
   pcl::PointCloud<pcl::PointXYZI> livoxPcl;
   int livoxCnt; //counts how many livox topics are merged in livoxPcl

   pcl::PointCloud<pcl::PointXYZI> rieglPcl; 
   bool rieglreceived;

   pcl::PointCloud<pcl::PointXYZI> ousterPcl;
   int ousterCnt;

   //keypoint
   Eigen::Vector3f keypoint;
   bool pointreceived;

   //subscriber
   ros::Subscriber livoxsub;
   ros::Subscriber rieglsub;
   ros::Subscriber oustersub;
   ros::Subscriber pointsub;

   //calibration
   Calibration livoxCal;
   Calibration rieglCal;
   Calibration ousterCal;

   std::string filepath;

public:
 RockIntensity()
 {

   //create an output filestream object
   std::string outpath, tag;
   n.getParam("outpath", outpath);
   n.getParam("tag", tag);
   bool bags;
   n.getParam("bags", bags);
   filepath = outpath + "/" + tag + "/";
   
   mkdir(filepath.c_str(), 0777); //make directory to save data in
   if(bags){ mkdir((filepath + "/bags/").c_str(), 0777);}
  
  //get calibration values from file 
  getCalibration();

  //livox subscribtion
  livoxCnt = 0;
  livoxsub = n.subscribe("livox/lidar", 1000, &RockIntensity::livoxCallback, this);

  //riegl subscribtion
  rieglreceived = false;
  rieglsub = n.subscribe("/riegl3d", 1000, &RockIntensity::rieglCallback, this);

  //ouster subscribtion
  ousterCnt = 0;
  oustersub = n.subscribe("/ouster/points", 1000, &RockIntensity::ousterCallback, this);
  
  //clicked point (rviz) subscription
  pointreceived = false;
  pointsub = n.subscribe("/clicked_point", 1000, &RockIntensity::pointCallback, this);

  ROS_INFO("[IntensityCorrection] subscribed to livox/lidar, /riegl3d, /ouster/points, and /clicked_point");

 }

 private:

/**
 * Function to retrieve values for calibration from file
**/
void getCalibration(){

   ROS_INFO("[IntensityCorrection] getting calibration values");

   //get filepath
   std::string path;
   n.getParam("calpath", path);
   ROS_INFO("[IntensityCorrection] got filepath: %s", path.c_str());

   //get calibration parameters for livox
   std::string fpath = path + "/livox.csv";
   std::ifstream file;
   file.open(fpath, std::ios::in);
   std::string line;
   std::getline(file, line); //discard header
   std::getline(file, line); //second line
   double white = std::stod(line);
   std::getline(file,line); //third line
   double black = std::stod(line);
   file.close();
   livoxCal.setValues(black, white);
   ROS_INFO("[IntensityCorrection] got livox calibration values: [%f,%f]", black, white);
   
   //get calibration parameters for riegl
   fpath = path + "/riegl.csv";
   file.open(fpath, std::ios::in);
   std::getline(file, line); //discard header
   std::getline(file, line); //second line = max value
   white = std::stod(line);
   std::getline(file,line); //third line
   black = std::stod(line);
   rieglCal.setValues(black, white);
   ROS_INFO("[IntensityCorrection] got riegl calibratoin range: [%f,%f]", black, white); 

   file.close();

   //get calibration parameters for ouster
   fpath = path + "/ouster.csv";
   file.open(fpath, std::ios::in);
   std::getline(file, line); //discard header
   std::getline(file, line); //second line = max value
   white = std::stod(line);
   std::getline(file,line); //third line
   black = std::stod(line);
   ousterCal.setValues(black, white);
   ROS_INFO("[IntensityCorrection] got ouster calibratoin range: [%f,%f]", black, white); 

   file.close();   
}

/**
 * Function to write the intensity data of all scanners to cvs file 
 * order of one line: LivoxIntensity, LivoxDeviation, LivoxWavelenght, OusterIntensity, OusterDeviation, OusterWavelenght, RieglIntensity, RieglDeviation, RieglWavelenght  
**/
void write_csv(std::string tag, std::vector<double> l_data, std::vector<double> o_data, std::vector<double> r_data){

   std::string target;
   n.getParam("target", target);
   filepath += target + ".csv";
   std::ofstream file(filepath, std::ios::out | std::ios::app);

   //write intensity data to file 
   file << l_data[0] << "," << l_data[1] << "," << l_lambda << ",";
   file << o_data[0] << "," << o_data[1] << "," << o_lambda << ",";
   file << r_data[0] << "," << r_data[1] << "," << r_lambda << "\n";
   file << "normalized \n";
   file << livoxCal.calibrate(l_data[0])  << ","  << l_lambda << ",";
   file << ousterCal.calibrate(o_data[0]) << ","  << o_lambda << ",";
   file << rieglCal.calibrate(r_data[0])  << ","  << r_lambda << "\n";

   ROS_INFO("[IntensityCorrection] Written data to '%s'", filepath.c_str());

   file.close();
   ros::shutdown(); //finished: kill ros node

}

/**
* Function gets called when new pointcloud topic message from riegl comes in
**/
void rieglCallback(const sensor_msgs::PointCloudConstPtr& msg)
{
   //convert to pcl poincloud (pointcloud to pointcloud2 to pcl)
   sensor_msgs::PointCloud2 temp;
    if (!sensor_msgs::convertPointCloudToPointCloud2 (*msg, temp))
      {
        ROS_ERROR ("[IntensityCorrection] Conversion from sensor_msgs::PointCloud to sensor_msgs::PointCloud2 failed!");
        return;
      }
   
   ROS_INFO("[IntensityCorrection] converted riegl pointcloud size : %ld", temp.data.size());
   pcl::PCLPointCloud2 pcl_pc2;
   pcl_conversions::toPCL(temp,pcl_pc2);
   ROS_INFO("[IntensityCorrection] converted riegl pointcloud size : %ld", pcl_pc2.data.size());
   pcl::PointCloud<pcl::PointXYZI> pcl;
   pcl::fromPCLPointCloud2(pcl_pc2, pcl);
   ROS_INFO("[IntensityCorrection] converted riegl pointcloud size : %ld", pcl.points.size());

   rieglPcl = pcl;
   rieglreceived = true;

}


/**
* Function gets called when new pointcloud topic message from ouster comes in, it adds the new published points to the class pointcloud of ouster
**/
void ousterCallback(const pcl::PointCloud<pcl::PointXYZI>::ConstPtr& pcl) //automatically transforms PointCloud2 to pcl cloud
{
   //add incoming points to existing pointcloud
   ousterPcl += *pcl;
   ousterCnt += 1;   
   
}

/**
* Function gets called when new pointcloud topic message from livox comes in, it adds the new published points to the class pointcloud of livox
* In case of the other laser scanner info being present already and enough livox points have been accomodated, intensity values get calculated
**/
void livoxCallback(const pcl::PointCloud<pcl::PointXYZI>::ConstPtr& pcl) 
{
   livoxPcl += *pcl;
   livoxCnt += 1;
   
   if (pointreceived && livoxCnt >= 50 && rieglreceived && ousterCnt >= 50){
      ROS_INFO("[IntensityCorrection] got all point clouds, starting calculation");
      calculate(); 
   }   
}

/**
 *  Get published point for livox
 **/
void pointCallback(const geometry_msgs::PointStampedConstPtr msg){
   ROS_INFO("[IntensityCorrection] got livoxpoint [%lf, %lf, %lf]", msg->point.x, msg->point.y, msg->point.z);
   keypoint(0) = msg->point.x;
   keypoint(1) = msg->point.y;
   keypoint(2) = msg->point.z;
   pointreceived = true;
}

 /**
 * Function uses pointclouds to calculate intensities
 **/
void calculate(){

   double radius;
   n.getParam("radius", radius);

   IntensityPoint L_Rock(keypoint, livoxPcl.makeShared(), radius, 0.01); 
   Eigen::Vector3f rpoint;// = rieglArun.transform(keypoint);
   Eigen::Vector3f opoint;// = ousterArun.transform(keypoint);

   //keypoint << 2.379,-0.125,-0.012;
   std::string oparam;
   n.getParam("opoint", oparam);
   double x, y, z;
   std::istringstream ostrm(oparam);
   ROS_INFO("[IntensotyPoint] received opoint = %s", oparam.c_str());
   ostrm >> x >> y >> z;
   opoint << x, y, z;

   std::string rparam;
   n.getParam("rpoint", rparam);
   std::istringstream rstrm(rparam);
   ROS_INFO("[IntensotyPoint] received rpoint = %s", rparam.c_str());
   rstrm >> x >> y >> z;
   rpoint << x, y, z;
   
   IntensityPoint R_Rock(rpoint, rieglPcl.makeShared(), radius, 0.01); 
   IntensityPoint O_Rock(opoint, ousterPcl.makeShared(), radius, 0.01);

   std::vector<double> livoxI  = L_Rock.getMeanI();
   std::vector<double> rieglI  = R_Rock.getMeanI();
   std::vector<double> ousterI = O_Rock.getMeanI();

   ROS_INFO("[IntensityCorrection] calculated livox  mean I [%lf+-%lf]", livoxI[0], livoxI[1]);
   ROS_INFO("[IntensityCorrection] calculated riegl  mean I [%lf+-%lf]", rieglI[0], rieglI[1]);
   ROS_INFO("[IntensityCorrection] calculated ouster mean I [%lf+-%lf]", ousterI[0], ousterI[1]);
   ROS_INFO("[IntensityCorrection] Normalized livox : [%f]", livoxCal.calibrate(livoxI[0]));
   ROS_INFO("[IntensityCorrection] Normalized riegl : [%f]", rieglCal.calibrate(rieglI[0]));
   ROS_INFO("[IntensityCorrection] Normalized ouster: [%f]", ousterCal.calibrate(ousterI[0]));

   std::string tag;
   n.getParam("tag", tag);
   ROS_INFO("[IntensityCorrection] writing csv"); //TODO: this does only print once, but 2 folders with each one csv file are written??
   write_csv(tag, livoxI, ousterI, rieglI);
}

};

int main(int argc, char* argv[])
{
   //initialize Ros node
   ros::init(argc, argv, "intensitycorrection_node");

   //printing and sending to rosout topic
   ROS_INFO("Intensity Correction running");

   //create object of class RockIntensity
   RockIntensity extractIntensity;

   //Dont exit automatically
   ros::spin();

}
